"""

This example demonstrates subclassing of Design_flute.

The resultant design can be used with Make_flute.

Note that Make_flute must be executed using this script,
so that it can unpickle the flute design.


Small warning:
The class hierarchy of instruments may change in future,
meaning you might need to inherit from a class with a slightly different name.

"""

import demakein, nesoni
from demakein import design, design_flute

class Design_pentatonic_flute(design_flute.Tapered_flute):
    # I like to give the designs as tenor-recorder sized
    # We'll default to actually making a smaller instrument
    transpose = 12
    
    
    n_holes = 5 # including embouchure hole
    
    # The fingerings need to include the open embouchure hole.
    fingerings = [
        ('D4', [1,1,1,1, 0]),
        ('E4', [0,1,1,1, 0]),
        ('G4', [0,0,1,1, 0]),
        ('A4', [0,0,0,1, 0]),
        ('B4', [0,0,0,0, 0]),
        ('D5', [1,1,1,1, 0]),
        ('E5', [0,1,1,1, 0]),
        ('G5', [0,0,1,1, 0]),
        ('A5', [0,0,0,1, 0]),
        ('B5', [0,0,0,0, 0]),
        ]

    initial_length = design.wavelength('D4') * 0.5
    
    # Holes 1, 2 and 3 should be fairly evenly spaced
    balance = [ None,    0.1,     None ]
    #Spacing of holes
    #           0,1,2    1,2,3    2,3,4      

    hole_horiz_angles = [ 0.0, 5.0, 0.0, 0.0,  0.0 ] 

    # Ways in which the flute may be jointed
    divisions = [ 
         # A single joint above hole 3
         [ (3, 0.0) ],
         
         # A joint halfway between hole 0 and hole 1,
         # and a joint above hole 3
         [ (0, 0.5), (3, 0.0) ],
         ]


if __name__ == '__main__':
    nesoni.run_toolbox(
        [ Design_pentatonic_flute, demakein.Make_flute ],
        show_make_flags=False)




from nesoni import config

from . import make, shape, profile

@config.help('Make a windcap for a reed instrument.')
@config.Float_flag('dock_diameter',
    'Diameter windcap should fit on top of.'
    )
@config.Float_flag('dock_length',
    'Length of socket.'
    )
class Make_windcap(make.Make):
    dock_diameter = 10.0
    dock_length = 15.0
    
    def run(self):
        wall = 2.0
        
        scale = self.dock_diameter
        
        z_dock_low   = 0.0
        z_dock_high  = self.dock_length
        z_top_low    = z_dock_high + scale * 2.5
        z_top_high   = z_top_low + wall
        z_pinch = z_top_low - scale*0.5
        z_mouth_high = z_top_high + scale * 0.75
        
        d_dock = self.dock_diameter
        d_inner = d_dock - wall
        d_outer = d_dock + wall
        d_dock_outer = d_outer + wall
        d_mouth_inner = wall
        d_mouth_outer = wall + wall*1.5
        
        outer_profile = profile.make_profile([
            (z_dock_low, d_dock_outer),
            (z_top_high-wall*0.5, d_outer),
            (z_top_high, d_outer-wall, d_mouth_outer+wall),
            (z_top_high+wall*0.5, d_mouth_outer),
            (z_mouth_high, d_mouth_outer),
            ])
        
        inner_profile = profile.make_profile([
            (z_dock_low, d_dock),
            (z_dock_high, d_dock, d_inner),
            (z_pinch, d_inner),
            (z_top_low, d_mouth_inner),
            (z_mouth_high, d_mouth_inner),
            ])
        
        stretch = (d_outer-d_mouth_outer)*0.5
        inner_stretch = profile.make_profile([
            (z_pinch, 0.0),
            (z_top_low, stretch)
            ])
        outer_stretch = profile.make_profile([
            (z_top_high, 0.0, stretch)
            ])
        
        cross_section = lambda d,s: shape.rounded_rectangle(-(d+s)*0.5,(d+s)*0.5,-d*0.5,d*0.5,d)
        
        outside = shape.extrude_profile(outer_profile, outer_stretch, cross_section=cross_section)
        inside = shape.extrude_profile(inner_profile, inner_stretch, cross_section=cross_section)
        
        thing = outside.copy()
        thing.remove(inside)
        self.save(thing, 'windcap')
        
        
        



import sys, os, math
sys.path.insert(0, os.path.split(__file__)[0])

import design, make, design_flute, profile, shape, pack

from nesoni import config

@config.help("""\
Produce 3D models using the output of "demakein design-*-flute:"
""")
@config.Bool_flag('open', 'Open both ends, thus requiring a cork.')
#@config.Bool_flag('mill', 'Create shapes for milling (rather than 3D printing).')
#@config.Float_flag('mill_diameter', 'Milling: Bit diameter for milling (affects gap size when packing pieces).')
#@config.Float_flag('mill_length', 'Milling: Wood length for milling.')
#@config.Float_flag('mill_width', 'Milling: Wood width for milling.')
#@config.Float_flag('mill_thickness', 'Milling: Wood thickness for milling.')
#@config.Int_flag('mill_scheme', 'Milling: Division scheme for milling.\nValid values: 2 3 4')
@config.Float_flag('emb_squareness', 'Squareness of embouchure hole, larger = squarer.')
@config.Float_flag('emb_aspect', 'Aspect ratio of embouchure hole, 1 = square, larger = wider.')
@config.Bool_flag('decorate', 'Add some decorations')
class Make_flute(make.Make_millable_instrument):
    #mill = False
    open = False

    #mill_diameter = 3.0    
    #mill_length = 180.0
    #mill_width = 130.0
    #mill_thickness = 19.0
    #mill_scheme = 4

    emb_squareness = 0.0
    emb_aspect = 1.5    
    decorate = False
    
    #SCHEMES = {
    #    2 : [
    #        [ 0.0, 0.6, 1.0 ],
    #        [ 0.0, 0.4, 1.0 ],
    #    ],
    #    
    #    3 : [
    #        [ 0.0, 0.3, 0.6, 1.0 ],
    #        [ 0.0, 0.4, 0.7, 1.0 ],
    #    ],
    #    
    #    4 : [
    #        [ 0.0, 0.2, 0.45, 0.7, 1.0 ],
    #        [ 0.0, 0.3, 0.55, 0.8, 1.0 ],
    #    ],
    #}
    
    def run(self):
        working = self.working
        designer = working.designer
        spec = working.spec
        workspace = self.get_workspace()
    
        length = spec.length * 1.05   # Extend a bit to allow cork.
        
        if self.open:
            inner_profile = spec.inner.clipped(-50,length+50)
            
            cork_length = length - spec.length
            cork_diameter = spec.inner(spec.length)
            with open(workspace/(self.prefix+'cork.txt'),'wt') as f:
                f.write('Cork length %.1fmm\n' % cork_length)
                f.write('Cork diameter %.1fmm\n' % cork_diameter)
        else:
            inner_profile = spec.inner.clipped(-50,spec.length)
        
        width = max(spec.outer.low)
        
        outer_profile = spec.outer.clipped(0,length)
        if self.decorate:
           emfrac = 1.0-spec.hole_positions[-1]/length
           for frac, align in [ (1.0-emfrac*2,1.0), (1.0,-1.0) ]:
               dpos = length * frac
               damount = outer_profile(dpos)*0.1
               dpos += damount * align
               deco_profile = profile.Profile(
                   [ dpos+damount*i for i in [-1,-0.333,0.333,1]],
                   [ damount*i      for i in [0,1,1,0] ],
               )
               outer_profile = outer_profile + deco_profile


        if self.emb_aspect > 1.0:
            emb_xpad = self.emb_squareness
            emb_ypad = (emb_xpad+1.0)*self.emb_aspect-1.0
        else:
            emb_ypad = self.emb_squareness
            emb_xpad = (emb_ypad+1.0)/self.emb_aspect-1.0

        self.make_instrument(
             inner_profile=inner_profile, outer_profile=outer_profile, 
             hole_positions=spec.hole_positions, hole_diameters=spec.hole_diameters, hole_vert_angles=spec.hole_angles, 
             #hole_horiz_angles=[0.0]*7,
             hole_horiz_angles=designer.hole_horiz_angles,
             xpad=[0.0]*(designer.n_holes-1)+[emb_xpad], 
             ypad=[0.0]*(designer.n_holes-1)+[emb_ypad],
             with_fingerpad=[True]*6+[False])
        
        #instrument = shape.extrude_profile(outer_profile)
        #outside = shape.extrude_profile(outer_profile)
        #
        #if self.open:        
        #    bore = shape.extrude_profile(spec.inner.clipped(-50,length+50))
        #else:
        #    bore = shape.extrude_profile(spec.inner.clipped(-50,spec.length))
        #
        ##xpad = [ .25,.25,0,0,0,0 ] + [ 0.0 ]
        ##ypad = [ .25,.25,0,0,0,0 ] + [ 0.75 ]
        #xpad = [ 0,0,0,0,0,0 ] + [ 0.0 ]
        #ypad = [ 0,0,0,0,0,0 ] + [ 0.75 ]
        #
        #
        #
        #for i, pos in enumerate(spec.hole_positions):
        #    angle = spec.hole_angles[i]
        #    radians = angle*math.pi/180.0
        #
        #    height = spec.outer(pos)*0.5 
        #    shift = math.sin(radians) * height
        #    
        #    hole_length = (
        #        math.sqrt(height*height+shift*shift) + 
        #        spec.hole_diameters[i]*0.5*abs(math.sin(radians)) + 
        #        4.0
        #    )
        #    hole = shape.prism(
        #        hole_length, spec.hole_diameters[i],
        #        shape.squared_circle(xpad[i], ypad[i]).with_effective_diameter
        #    )
        #    hole.rotate(1,0,0, -90-angle)
        #    hole.move(0,0,pos + shift)
        #    bore.add(hole)
        #    if angle:
        #        outside.remove(hole)
        #
        #instrument.remove(bore)    
        #instrument.rotate(0,0,1, 180)
        #self.save(instrument,'instrument')
        
        self.make_parts(up = False)
        
        #if self.mill:        
        #    shapes = pack.cut_and_pack(
        #        working.outside, working.bore,
        #        self.SCHEMES[self.mill_scheme][0], self.SCHEMES[self.mill_scheme][1],
        #        xsize=self.mill_length, 
        #        ysize=self.mill_width, 
        #        zsize=self.mill_thickness,
        #        bit_diameter=self.mill_diameter,
        #        save=self.save,
        #    )
        #
        #else:
        #    for division in designer.divisions:
        #        cuts = [ ]
        #        for hole, above in division:
        #            if hole >= 0:
        #                cut = spec.hole_positions[hole] + 2*spec.hole_diameters[hole]
        #            else:
        #                cut = 0.0
        #            cut += (length-cut)*above
        #            cuts.append(cut)
        #            
        #        self.segment(cuts, up=False)


            #cut1 = min(spec.hole_positions[3],spec.inner_hole_positions[3])-spec.hole_diameters[3] * 0.75
            #cut1 -= spec.inner(cut1)*0.5
            ##cut1 = spec.inner_hole_positions[2]*0.5+spec.inner_hole_positions[3] * 0.5
            ##cut2 = cut1 + (length-cut1)*0.3 + spec.outer.maximum()*0.7
            #cut2 = length * 0.62
            #
            #self.segment([ cut1, cut2 ], length, up=False)
            
    

if __name__ == '__main__': 
    shape.main_action(Make_flute())




"""

Matrices at arbitrary offsets

"""

import numpy, subprocess, os

def bound_intersection(bounds1, bounds2):
    (x1,y1,width1,height1) = bounds1
    (x2,y2,width2,height2) = bounds2
    if width1 == 0 or height1 == 0: return bounds2
    if width2 == 0 or height2 == 0: return bounds1
    x = max(x1,x2)
    y = max(y1,y2)
    width = max(0, min(x1+width1,x2+width2)-x)
    height = max(0, min(y1+height1,y2+height2)-y)
    return (x,y,width,height)

def bound_union(bounds1, bounds2):
    (x1,y1,width1,height1) = bounds1
    (x2,y2,width2,height2) = bounds2
    if width1 == 0 or height1 == 0: return bounds2
    if width2 == 0 or height2 == 0: return bounds1
    x = min(x1,x2)
    y = min(y1,y2)
    width = max(x1+width1,x2+width2)-x
    height = max(y1+height1,y2+height2)-y
    return (x,y,width,height)

class Big_matrix(object):
    def __init__(self, x,y,data):
        self.x = x
        self.y = y
        self.height, self.width = data.shape[:2]
        self.data = data
    
    def shift(self, x,y):
        return Big_matrix(self.x+x,self.y+y,self.data)
    
    def bounds(self):
        return (self.x,self.y,self.width,self.height)
    
    def clip(self, x,y,width,height):
        if y == self.y and x == self.x and width == self.width and height == self.height: 
            return self
        result = zeros(x,y,width,height,self.data.dtype)
        x1 = max(self.x,x)
        x2 = min(self.x+self.width,x+width)
        y1 = max(self.y,y)
        y2 = min(self.y+self.height,y+height)
        if y1 < y2 and x1 < x2:
            result.data[y1-y:y2-y,x1-x:x2-x] = self.data[y1-self.y:y2-self.y,x1-self.x:x2-self.x]
        return result
    
    def apply(self, other, bounds, operation):
        a = self.clip(*bounds)
        b = other.clip(*bounds)
        return Big_matrix(a.x,a.y,operation(a.data,b.data))
    
    def union_apply(self, other, operation):
        bounds = bound_union(self.bounds(),other.bounds())
        return self.apply(other, bounds, operation)
        
    def intersection_apply(self, other, operation):
        bounds = bound_intersection(self.bounds(),other.bounds())
        return self.apply(other, bounds, operation)
    
    def __and__(self,other):
        return self.intersection_apply(other,lambda a,b: a&b)
    def __or__(self,other):
        return self.union_apply(other,lambda a,b: a|b)
    
    def and_not(self, other):
        return self.apply(other, self.bounds(), lambda a,b: a&~b)
    
    def spans(self):
        for y in range(self.height):
            start = 0
            for x in range(self.width):
                if not self.data[y,x]:
                    if start != x:
                        yield y+self.y, start+self.x, x+self.x
                    start = x+1
            if start != self.width:
                yield y+self.y, start+self.x, self.width+self.x

    def morph(self, mask, operation):
        spans = [ None, self ]
        def get_span(n):
            while len(spans) <= n:
                spans.append(operation(spans[-1],self.shift(len(spans)-1,0)))
            return spans[n]
        result = None
        for y,x1,x2 in mask.spans():
            span = get_span(x2-x1).shift(x1,y)
            if result is None:
                result = span
            else:
                result = operation(result, span)
        assert result is not None, 'empty mask'
        return result
    
    def dilate(self,mask):
        return self.morph(mask, lambda a,b: a|b)
    
    def erode(self,mask):
        return self.morph(mask, lambda a,b: a&b)
    
    def open(self,mask):
        return self.erode(mask).dilate(mask)
    
    def close(self,mask):
        return self.dilate(mask).erode(mask)

    def trace(self,res=1.0):
        return trace(self,res)


def ones(x,y,width,height,type=bool):
    return Big_matrix(x,y, numpy.ones((height,width),type))

def zeros(x,y,width,height,type=bool):
    return Big_matrix(x,y, numpy.zeros((height,width),type))

def zero(type=bool):
    return zeros(0,0,0,0,type)






def line_param(x1,y1,x2,y2):
    if x2 == x1:
        return 0.0,y1
    m = (y2-y1)/(x2-x1)
    c = y1-m*x1
    return m,c

def int_floor(x):
    return int(numpy.floor(x))

def int_ceil(x):
    return int(numpy.ceil(x))

def make_mask(lines):
    base_x = int_floor(min( min(x1,x2) for x1,y1,x2,y2 in lines ))
    base_y = int_floor(min( min(y1,y2) for x1,y1,x2,y2 in lines ))
    width  = int_ceil(max( max(x1,x2) for x1,y1,x2,y2 in lines )) - base_x + 1
    height = int_ceil(max( max(y1,y2) for x1,y1,x2,y2 in lines )) - base_y + 1
    
    line_count = { }
    for x1,y1,x2,y2 in lines:
        if x1 < x2:
            n = 1
            key = (x1,y1,x2,y2)
        else:
            n = -1
            key = (x2,y2,x1,y1)
        line_count[key] = line_count.get(key,0) + n
    
    count = numpy.zeros((height,width), 'int32')
    
    for (x1,y1,x2,y2), n in line_count.items():
        if n == 0: continue 
        ix1 = int_ceil(x1)
        ix2 = int_ceil(x2)
        if ix1 == ix2: continue

        m,c = line_param(x1,y1,x2,y2)
        for x in range(ix1,ix2):
            y = int_ceil(m*x+c)
            count[y-base_y,x-base_x] += n
    
    for y in range(1, height):
        count[y] += count[y-1]

    return Big_matrix(base_x, base_y, count > 0)




def write_pbm(f, data):
    f.write('P1\n%d %d\n' % (data.shape[1],data.shape[0]))
    for line in data:
        f.write(''.join([ '1' if item else '0' for item in line ]))
        #f.write('\n')

def save(prefix, mask):
    with open(prefix+'.pbm','wt') as f: write_pbm(f, mask.data)
    os.system('pnmtopng <%s.pbm >%s.png' % (prefix,prefix))


def trace(mask, res=1.0):
    import shape
    
    process = subprocess.Popen([
        'potrace', '-a', '-1', '-t', '0', '-u', '100', '-b', 'svg'
    ], stdin=subprocess.PIPE, stdout=subprocess.PIPE, universal_newlines=True)
    
    write_pbm(process.stdin, mask.data)

    from xml.etree import ElementTree
    doc = ElementTree.parse(process.stdout)

    loops = [ ]    
    for path in doc.getroot().findall('{http://www.w3.org/2000/svg}g/{http://www.w3.org/2000/svg}path'):
        items = path.attrib['d'].split()
        items = [ int(item.strip('Mczl')) for item in items ]
        for i in range(2,len(items)):
            items[i] += items[i-2]
        
        loop = [ ]
        for i in range(len(items)//2):
            loop.append( (mask.x+items[i*2]*0.01-0.5,mask.y+mask.data.shape[0]-items[i*2+1]*0.01-0.5) )
        
        i = 0
        while i < len(loop):
            if loop[i] == loop[(i+1)%len(loop)]:
                del loop[i]
            else:
                i += 1
        
        assert len(loop) == len(set(loop))        
        
        loops.append( shape.Loop(loop[::-1]).scale(1.0/res) )
    return loops
    
    #loop = [ ]
    #loops = [ loop ]
    #for line in process.stdout:
    #     parts = line.rstrip().split()
    #     if parts[0] != 'TYPE:': continue
    #     assert parts[2] == 'X:' and parts[4] == 'Y:'
    #     if parts[1] == '2': continue
    #     if parts[1] == '3':
    #          loop = [ ]
    #          loops.append(loop)
    #     loop.append( (0.5+mask.x+int(parts[3])*0.01,0.5+mask.y+int(parts[5])*0.01) )
    #return [ shape.Loop(item[::-1]) for item in loops ]


if __name__ == '__main__':
    import shape
    loop = shape.Loop( shape.circle(50.0)[:50] )
    with open('/tmp/test.pbm','wt') as f: write_pbm(f,loop.mask(1.0).data)
    print( max(x for x,y in loop) - min(x for x,y in loop) )
    for i in range(10):
         mask = loop.mask(1.0)
         with open('/tmp/test2.pbm','wt') as f: write_pbm(f,mask.data)
         [ loop ] = trace(mask)
         print( max(x for x,y in loop), min(x for x,y in loop) )
         print( max(y for x,y in loop), min(y for x,y in loop) )
         print()
         #print( min(x for x,y in loop) )
    #mask[5,5] = 0
    with open('/tmp/test2.pbm','wt') as f: write_pbm(f,mask.data)
    
    
    #mask = shape.circle(20).mask(40)
    #
    #thing = zeros(0,0,100,100)
    #thing.data[:,:] = True
    #
    #thing = thing | thing.shift(100,100)
    #thing = thing.close(mask)
    #with open('/tmp/test3.pbm','wt') as f: write_pbm(f,thing.data)
    
    





import math, copy

from nesoni import config

from . import design, profile

@config.Float_flag('tweak_gapextra')
#@config.Float_flag('tweak_bulgediameter')
#@config.Float_flag('tweak_bulgepos1')
#@config.Float_flag('tweak_bulgepos2')
@config.Float_flag('tweak_boreless')
class Design_whistle(design.Instrument_designer_with_bore_scale):
    def get_whistle_maker(self):
        from . import make_whistle
        bore = self.inner_diameters[-1]
        outside = self.outer_diameters[-1]
        if self.outer_add: outside += bore
        return make_whistle.Make_whistle_head(
            bore = bore,
            outside = outside,
            )
        
    closed_top = False
    
    divisions = [ ]

    # For gap_width 0.5, gap_length 0.25, from soprano
    # tweak_gapextra = 0.37
    
    # From soprano recorder
    tweak_gapextra = 0.75 #0.71
    #tweak_bulgediameter = 1.0
    #tweak_bulgepos1 = 0.5
    #tweak_bulgepos2 = 0.9
    tweak_boreless = 0.65 #0.49
    
    xpad = 0.0
    ypad = 0.0
    
    def patch_instrument(self, inst):
        inst = copy.copy(inst)
        inst.true_length = inst.length
        inst.true_inner = inst.inner
        
        bore_diameter = inst.inner(inst.length)
        inst.length -= bore_diameter * self.tweak_boreless
        inst.inner = inst.inner.clipped(0.0,inst.length)

        #bulge_pos1 = self.tweak_bulgepos1 * inst.length
        #bulge_pos2 = self.tweak_bulgepos2 * inst.length
        #if bulge_pos1 > bulge_pos2:
        #    bulge_pos1, bulge_pos2 = bulge_pos2, bulge_pos1
        #bulge_amount = (self.tweak_bulgediameter-1.0) * bore_diameter
        #inst.inner = inst.inner + profile.Profile(
        #    [ bulge_pos1, bulge_pos2 ],
        #    [ 0.0, bulge_amount ],
        #    [ bulge_amount, 0.0 ],
        #    )
        
        maker = self.get_whistle_maker()
        diameter = maker.effective_gap_diameter()             
        length = maker.effective_gap_height() / 2.0
        #        ^ The gap is this high on one side and a blade on the other side
        #          split the difference.
        length += diameter * self.tweak_gapextra
        
        inst.inner = profile.Profile(
            inst.inner.pos + [ inst.length+length ],
            inst.inner.low + [ diameter ],
            inst.inner.high[:-1] + [ diameter, diameter ],
            )
        inst.length += length
        return inst


@config.help("""\
Design a whistle with pennywhistle fingering.
""")
class Design_folk_whistle(Design_whistle):    
    transpose = 12
    
    divisions = [
        [(5,0)],
        [(1,0),(5,0)],
        [(1,0),(5,0),(5,0.5)],
        ]
    
    initial_length = design.wavelength('D4') * 0.5
    
    min_hole_diameters = design.bore_scaler([ 3.0 ]*6)
    max_hole_diameters = design.bore_scaler([ 12.0 ]*6)
    
    horiz_angles = [ 0.0 ] * 6
    #mid_cut = 3

    #balance = [ 0.05, 0.05, 0.05, 0.05 ]
    balance = [ 0.05, None,None, 0.05 ]
    
    min_hole_spacing = design.scaler([ None, None, 35.0, None, None ])
    max_hole_spacing = design.sqrt_scaler([ 35.0, 35.0, None, 35.0, 35.0 ])
    
    #min_hole_spacing = [ 12.0 ] * 5
    #@property
    #def min_hole_spacing(self):
    #    b = self.inner_diameters[-1]
    #    return [ b*1.15 ] * 5 #[ b,b,b*1.25,b,b ]

    inner_diameters = design.bore_scaler([ 14.0, 14.0, 20.0, 22.0, 22.0, 20.0, 20.0 ])
    initial_inner_fractions = [ 0.2, 0.6,0.65,0.7,0.75 ]
    min_inner_fraction_sep = [ 0.01, 0.5 ] + [ 0.01 ] * 4

#    outer_diameters = design.bore_scaler([ 38.0, 28.0, 28.0 ])    
#    initial_outer_fractions = [ 0.15 ]
#    min_outer_fraction_sep = [ 0.15, 0.8 ]

    outer_diameters = design.bore_scaler([ 40.0, 28.0, 28.0, 32.0, 32.0 ])    
    outer_angles = [ -15.0, (0.0,None), None, None, None ]
    initial_outer_fractions = [ 0.15, 0.5, 0.85 ]
    min_outer_fraction_sep = [ 0.15, 0.3, 0.35, 0.15 ]
    
    #min_outer_fraction_sep = [ 0.01, 0.5 ]

    fingerings = [
        ('D4',   [1,1,1,1,1,1]),
        ('E4',   [0,1,1,1,1,1]),
        ('F#4',  [0,0,1,1,1,1]),
        ('G4',   [0,0,0,1,1,1]),
        ('A4',   [0,0,0,0,1,1]),
        ('B4',   [0,0,0,0,0,1]),
        ('C5',   [0,0,0,1,1,0]),
        ('C#5',  [0,0,0,0,0,0]),
        ('D5',   [1,1,1,1,1,0]),
        ('E5',   [0,1,1,1,1,1]),
        ('F#5',  [0,0,1,1,1,1]),
        ('G5',   [0,0,0,1,1,1]),
        ('A5',   [0,0,0,0,1,1]),
        ('B5',   [0,0,0,0,0,1]),
        #('C#6',  [1,1,1,0,0,0]),
        #('C#6',  [0,0,0,0,0,0]),
        ('D6',   [1,1,1,1,1,1]),
    ]


@config.help("""\
Design a recorder. \
This is like a penny-whistle, but with a complicated fingering system.
""")
class Design_recorder(Design_whistle):
    transpose = 12
    
    divisions = [
        [(7,0)],
        [(0,0),(7,0)],
        [(0,0),(3,0),(7,0)],
        [(0,0),(3,0),(7,0),(7,0.5)],
        [(0,0),(2,0),(4,0),(7,0),(7,0.5)],
        ]
    
    initial_length = design.wavelength('C4') * 0.5
    
    min_hole_diameters = design.bore_scaler([ 3.0 ]*8)
    max_hole_diameters = design.bore_scaler([ 12.0 ] + [ 14.0 ]*7)
    min_hole_spacing = design.sqrt_scaler([ 0.0 ]*6+[-50.0])

    horiz_angles = [ -15.0 ] + [ 0.0 ] * 6 + [ 180.0 ]
    hole_angles = [ -30, 30, -30, 30, 0, 0, 0, 0 ]

    balance = [ 0.1, 0.05, None, None, 0.05, None ]
    #balance = [ 0.2 ]*4
    #@property
    #def min_hole_spacing(self):
    #    b = self.inner_diameters[-1]
    #    return [ b*1.15 ] * 5 #[ b,b,b*1.25,b,b ]

#    inner_diameters = design.bore_scaler([ 17.0, 20.0, 20.0, 23.0, 20.0, 20.0 ])
    inner_diameters = design.bore_scaler([ 20.0, 20.0, 19.0, 23.0, 23.0, 20.0, 20.0 ])
    initial_inner_fractions = [ 0.6,0.65,0.7,0.75,0.8 ]
    min_inner_fraction_sep = [ 0.3,0.01,0.01,0.01,0.01,0.01 ]

    outer_diameters = design.bore_scaler([ 40.0, 28.0, 28.0, 32.0, 32.0 ])    
    initial_outer_fractions = [ 0.15, 0.6, 0.85 ]
    min_outer_fraction_sep = [ 0.15, 0.3, 0.35, 0.15 ]


    fingerings = [
        ('C4',   [1,1,1,1,1,1,1,1]),
        # Inter-register locking
        ('C5',   [1,1,1,1,1,1,1,1]),
        ('G5',   [1,1,1,1,1,1,1,1]),

        ('D4',   [0,1,1,1,1,1,1,1]),
        # Inter-register locking
        ('D5',   [0,1,1,1,1,1,1,1]),
        ('A5',   [0,1,1,1,1,1,1,1]),

        ('E4',   [0,0,1,1,1,1,1,1]),
        # Inter-register locking
        ('B5',   [0,0,1,1,1,1,1,1]),
        
        #Tendency to be sharp, due to inter register locking?
        ('F4',   [1,1,0,1,1,1,1,1]),
        ('F4',   [0,1,0,1,1,1,1,1]),
        
        ('F#4',  [0,1,1,0,1,1,1,1]),
        ('G4',   [0,0,0,0,1,1,1,1]),
        ('G#4',  [0,1,1,1,0,1,1,1]),
        ('A4',   [0,0,0,0,0,1,1,1]),
        ('Bb4',  [0,0,0,1,1,0,1,1]),
        ('B4',   [0,0,0,0,0,0,1,1]),
        ('C5',   [0,0,0,0,0,1,0,1]),
        ('C#5',  [0,0,0,0,0,1,1,0]),
        ('D5',   [0,0,0,0,0,1,0,0]),


        ('Eb5',  [0,1,1,1,1,1,0,0]),
        
        ('E5',   [0,0,1,1,1,1,1,0]),
        ('E5',   [0,0,1,1,1,1,1,1]),
        
        ('F5',   [0,1,0,1,1,1,1,0]),
        ('F5',   [0,1,0,1,1,1,1,1]),
        
        ('F#5',  [0,0,1,0,1,1,1,0]),
        ('F#5',  [0,0,1,0,1,1,1,1]),

        #('G5',   [0,0,0,0,1,1,1,0]),
        ('G5',   [0,0,0,0,1,1,1,1]),

        #('G#5',  [0,0,0,1,0,1,1,1]),
        ('A5',   [0,0,0,0,0,1,1,1]),
#        ('Bb5',  [0,1,1,1,0,1,1,1]),
#        ('B5',   [0,0,1,1,0,1,1,1]),
    ]
    

@config.help("""\
Design a three hole pipe, i.e. the pipe in pipe-and-tabor.
""")
class Design_three_hole_whistle(Design_whistle):
    transpose = 24
    
    bore_scale = 1.2
    
    divisions = [
        [],
        [(2,0.3)],
        [(2,0.05),(2,0.5)],
        [(2,0.05),(2,0.35),(2,0.65)],
        [(1,0.5),(2,0.2),(2,0.45),(2,0.7)],
        ]
    
    xpad = 0.0
    ypad = 1.0
    
    initial_length = design.wavelength('D3') * 0.5
    
    min_hole_diameters = design.bore_scaler([ 3.0 ]*3)
    max_hole_diameters = design.bore_scaler([ 14.0 ]*3)
    max_hole_spacing = design.scaler([ 100.0, 100.0 ])
    top_clearance_fraction = 0.6 #Avoid local minimum with holes in center of instrument
    
    hole_angles = [ -30.0, 30.0, 30.0 ]
    horiz_angles = [ 0.0, 0.0, 180.0 ]
    mid_cut = 2
    
    #min_hole_spacing = [ -20.0, -20.0 ]
    
    initial_hole_fractions = [ 0.1,0.15,0.2 ]
    
    inner_diameters = design.bore_scaler([ 10.0, 10.0, 12.5, 15.0, 17.5, 20.0,  20.0 ])
    initial_inner_fractions = [ 0.1, 0.2, 0.3, 0.4, 0.5 ]
    
    outer_diameters = design.bore_scaler([ 45.0, 32.0, 32.0 ])
    outer_angles = [ -25.0, (0.0,None), None ]
    min_outer_fraction_sep = [ 0.095, 0.0 ]
    max_outer_fraction_sep = [ 0.105, 1.0 ]
    initial_outer_fractions = [ 0.1 ]
        
    min_inner_fraction_sep = [ 0.01 ] * 8

    fingerings = [
        ('D3', [1,1,1]),
        ('E3', [0,1,1]),
        ('F#3',[0,0,1]),
        ('G3', [0,0,0]),
    
        ('D4', [1,1,1]),
        ('E4', [0,1,1]),
        ('F#4',[0,0,1]),
        ('G4', [0,0,0]),
        ('A4', [1,1,1]),
        ('B4', [0,1,1]),
        ('C#5',[0,0,1]),
        ('D5', [0,0,0]),
        ('D5', [1,1,1]),
        ('E5', [0,1,1]),
        ('F#5',[0,0,1]),
        ('G5', [0,0,0]),
        ]






import demakein, nesoni

class Design_simple_shawm(demakein.design.Instrument_designer):
    closed_top = True

    # Bore will be 35mm at end, 4mm at top
    inner_diameters = [ 35.0, 4.0 ]

    # Outer profile to be in addition to inner profile    
    outer_add = True
    # This determines the depth of the holes, 5mm extra diameter = 2.5mm depth
    outer_diameters = [ 5.0, 5.0 ]
    
    # Limits on finger holes sizes
    min_hole_diameters = [ 2.0 ]*6
    max_hole_diameters = [ 15.0 ]*6
    
    # Limit how close successive holes can be to each other
    #min_hole_spacing = [ 5,5,5,20,20 ]
    # see also max_hole_spacing
    #          balance
        
    # Very simple fingering system
    initial_length = demakein.design.wavelength('D4') * 0.4
    fingerings = [
        ('D4',  [1,1,1,1,1,1]),
        ('E4',  [0,1,1,1,1,1]),
        ('F#4', [0,0,1,1,1,1]),
        ('G4',  [0,0,0,1,1,1]),
        ('A4',  [0,0,0,0,1,1]),
        ('B4',  [0,0,0,0,0,1]),
        ('C#5', [0,0,0,0,0,0]),
        ('D5',  [1,1,1,1,1,1]),
        ('E5',  [0,1,1,1,1,1]),
        ('F#5', [0,0,1,1,1,1]),
        ('G5',  [0,0,0,1,1,1]),
        ('A5',  [0,0,0,0,1,1]),
        ('B5',  [0,0,0,0,0,1]),
        ('C#6', [0,0,0,0,0,0]),
    ]    

if __name__ == '__main__':
    nesoni.run_tool(Design_simple_shawm)
    
    # or
    #nesoni.run_toolbox([ 
    #    Design_simple_shawm,
    #    #...
    #    ])



import math

from nesoni import config

from . import make, shape, profile

def wobble(diameter=1.0, wobble=0.5, spin=0.0, period=16, n=256):
    radius = diameter * 0.5
    result = [ ]
    for i in range(n):
        a = (i+0.5)*math.pi*2.0/n 
        s = spin*math.pi / 180.0
        b = math.cos((a+s)*period)
        c = math.cos((a-s*0.5)*period)*0.75
        d = 1-(1-b)*(1-c)
        r2 = radius*(1.0 + wobble*( d -0.5) )
        result.append( (math.cos(a)*r2, math.sin(a)*r2) )
    return shape.Loop(result)


@config.help(
'Make a gratuitous bauble.',
"""\
This tool is used when you use "make-shawm: --bauble yes".

Make your own gratuitous baubles to fit the cylindrical object of your choice.
"""
)
@config.Float_flag('dock_diameter','Diameter of cylinder to fit.')
@config.Float_flag('dock_length','Length of cylinder to fit.')
class Make_bauble(make.Make):
    dock_diameter = 40.0
    dock_length = 5.0
    
    def run(self):
        length = self.dock_diameter * 1.5
        inlength = length * 0.9
        pos_outer = [ 0.0, length ]
        diam_outer = [ self.dock_diameter+2.0, 0.0 ]
        angle = [ 20.0, -10.0 ]
        p_outer = profile.curved_profile(
            pos_outer, diam_outer, diam_outer, angle, angle)
        
        pos_inner = [ 0.0, inlength ]
        diam_inner = [ self.dock_diameter, 0.0 ]
        p_inner = profile.curved_profile(
            pos_inner, diam_inner, diam_inner, angle, angle)
        
        spin = profile.make_profile([(0.0,0.0),(length, 120.0)])
        
        wob = profile.make_profile([(self.dock_length*0.5,0.0),(self.dock_length,1.0),(length,0.0)])
        
        bauble = shape.extrude_profile(p_outer,spin,wob, cross_section=lambda d,s,w: wobble(d,w*0.1,s,12))        
        inside = shape.extrude_profile(
            p_inner.clipped(self.dock_length+1.0,inlength),spin.clipped(self.dock_length+1.0,inlength),wob.clipped(self.dock_length+1.0,inlength), 
            cross_section=lambda d,s,w: wobble(d,w*-0.1,s,12))
        bauble.remove(inside)
        
        dock_profile = profile.make_profile([(0.0,self.dock_diameter),(self.dock_length,self.dock_diameter),(self.dock_length+self.dock_diameter*0.5,0.0)])
        dock = shape.extrude_profile(dock_profile)
        #bauble.add(shape.extrude_profile(p_outer.clipped(0.0,self.dock_length+1.0)))
        bauble.remove(dock)
        
        self.save(bauble, 'bauble')
        return bauble




import demakein

import nesoni
from nesoni import config

@config.help("""\
Design and make all instruments, in a variety of sizes.
""","""\
Note that I use the terms "soprano", "alto", "tenor", and "bass" to \
refer to the sizes of instruments. Flutes and shawms I name this way are actually \
an octave above the singing voices of the same name.
""")
@config.Bool_flag('panpipes', 'Do panpipes.')
@config.Bool_flag('flutes', 'Do flutes.')
@config.Bool_flag('whistles', 'Do whistles.')
@config.Bool_flag('shawms', 'Do shawms.')
@config.String_flag('version', 'version code, for file names')
class All(config.Action_with_output_dir):
    panpipes = True
    flutes = True
    whistles = True
    shawms = True
    version = 'v'+demakein.VERSION.lstrip('0.')
    
    def _do_flute(self, model_name, model_code, size_name, size_code, designer, transpose):
        workspace = self.get_workspace()
        outdir = workspace / (model_name+'-'+size_name)

        designer(
            outdir,
            transpose=transpose
            ).make()
        
        demakein.Make_flute(
            outdir,
            prefix = model_code+size_code+'-'+self.version+'-',
            decorate=True,
            ).make()

    def _do_shawm(self, model_name, model_code, size_name, size_code, designer, transpose, bore):
        workspace = self.get_workspace()
        outdir = workspace / (model_name+'-'+size_name)
    
        designer(
            outdir,
            transpose=transpose,
            bore=bore,
            ).make()

        demakein.Make_shawm(
            outdir,
            prefix=model_code+size_code+'-'+self.version+'-',
            decorate=True,
            ).make()
    
    def _do_folk_whistle(self, model_name, model_code, size_name, size_code, transpose):
        workspace = self.get_workspace()
        outdir = workspace / (model_name+'-'+size_name)
        
        demakein.Design_folk_whistle(
            outdir,
            transpose=transpose,
            ).make()
        
        demakein.Make_whistle(
            outdir,
            prefix=model_code+size_code+'-'+self.version+'-',
            ).make()
    
    def run(self):
        workspace = self.get_workspace()
        
        stage = nesoni.Stage()
        
        if self.panpipes:
            if self.make:
                demakein.Make_panpipe(
                    workspace/'panpipe'
                    ).process_make(stage)
        
        if self.flutes:
            for model_name, model_code, designer in [
                    ('folk-flute-straight', 'FFS', demakein.Design_straight_folk_flute),
                    ('folk-flute-tapered',  'FFT', demakein.Design_tapered_folk_flute),
                    ('pflute-straight',     'PFS', demakein.Design_straight_pflute),
                    ('pflute-tapered',      'PFT', demakein.Design_tapered_pflute),
                    ]:
                for size_name, size_code, transpose in [
                        ('tenor', 't', 0),
                        ('alto', 'a', 5),
                        ('soprano', 's', 12),
                        ]:
                    stage.process(self._do_flute,model_name,model_code,size_name,size_code,designer,transpose)

        if self.whistles:
            for model_name, model_code in [
                    ('folk-whistle', 'FW'),
                    ]:
                for size_name, size_code, transpose in [
                        ('tenor', 't', 0),
                        ('alto', 'a', 5),
                        ('soprano', 's', 12),
                        ('sopranino', 'ss', 17),
                        ]:
                    stage.process(self._do_folk_whistle,model_name,model_code,size_name,size_code,transpose)

        if self.shawms:
            for model_name, model_code, designer in [
                 ('shawm', 'SH', demakein.Design_shawm),
                 ('folk-shawm', 'FSH', demakein.Design_folk_shawm),
                 ]:
                 for size_name, size_code, transpose, bore in [
                         ('4mm-alto', '4a', 5, 4.0),
                         ('4mm-tenor', '4t', 0, 4.0),
                         ('6mm-tenor', '6t', 0, 6.0),
                         ('6mm-bass', '6b', -7, 6.0),
                         ]:
                     stage.process(self._do_shawm,model_name,model_code,size_name,size_code,designer,transpose,bore)

        stage.barrier()





"""

8mm thickness, 6mm holes

C6

D6 
got: 1120Hz -> 309mm
want: 294.6
error: 14.4mm

E6 
got: 1231Hz -> 281.2mm
want: 262.5mm
error: 18.7mm

G6 
got: 1462 -> 236.7mm
want: 220.7mm
error: 16.0mm

C7 
got: 1942Hz ~ 178.2mm
want: 165.4mm
error: 12.8mm

average: 15.5 wavelength = 3.87 lengths = 0.645 diameters

"""



import sys, os, math
sys.path.insert(0, os.path.split(__file__)[0])

import design, make, shape, profile

from nesoni import config

RES = 20

LENGTH_CORRECTION = -0.7
   #-0.645 # * diameter
   # Hole is rounded at end, so make it a little longer
   # - unknown further correction 

SCALE = [
    'C6',
    'D6',
    'E6',
    'F6',
    'G6',
    'A6',
    'B6',
    'C7',
    'D7',
]

#SCALE = [
#    'F6',
#    'G6',
#    'A6',
#    'C7',
#    'D7',
#    'F7',
#]

def make_hole(diameter, length):
    pos = [ 0.0 ]
    diam = [ diameter ]
    
    radius = diameter*0.5
    n = 4
    for i in range(0,n):
        x = math.pi/2 * float(i)/n
        pos.append(length+radius*(math.sin(x)-1))
        diam.append(diameter*math.cos(x))
    
    prof = profile.Profile(pos,diam,diam)
    hole = shape.extrude_profile(prof)
    
    return hole

@config.help("""\
Make a Viking-style panpipe.
""")
@config.Float_flag('thickness', 'Instrument thickness.\n(Wood thickness should be half this if milling.)')
@config.Float_flag('wall', 'Minimum wall thickness')
@config.Int_flag('transpose', 'Transpose (semitones)')
class Make_panpipe(make.Make):
    thickness = 8.0
    wall = 1.0
    transpose = 0
    
    def run(self):
        zsize = self.thickness
        
        pad = self.wall
        diameter = zsize - pad*2
        
        self.log.log('Thickness: %.1fmm\n' % zsize)
        self.log.log('Min wall thickness: %.1fmm\n' % pad)
        self.log.log('Hole diameter: %.1fmm\n' % diameter)
        
        negatives = [ ]
        lengths = [ ]
        xs = [ ]
        for i, note in enumerate(SCALE):
            print 'Make hole %d / %d' % (i+1,len(SCALE))
            length = design.wavelength(note,self.transpose)*0.25 + LENGTH_CORRECTION*diameter
            x = i*(diameter+pad)
            lengths.append(length)
            xs.append(x)
            hole = make_hole(diameter, length)
            hole.move(x,0,0)
            negatives.append(hole)
        
        string_hole_diameter = diameter*1.5
        string_hole_loop = shape.circle(string_hole_diameter)
        string_hole = shape.extrusion([-zsize,zsize],[string_hole_loop,string_hole_loop])
        string_hole.rotate(1,0,0, 90)
        
        xlow = xs[0]-zsize
        #xmid = xs[-1]*0.5
        xhigh = xs[-1]+zsize
        
        zhigh = lengths[0]+zsize*0.5
        #zmid  = max(lengths[-1]+zsize*0.5, zhigh-(xhigh-xmid))
        
        trim = min(xhigh-xlow,zhigh) * 0.5
        
        #string_x = (xmid+xhigh)*0.5 - diameter
        #string_z = (zmid+zhigh)*0.5 - diameter
        #string_z = lengths[3] + diameter*2
        #string_x = xhigh-trim*0.5-diameter
        #string_z = zhigh-trim*0.5-diameter
        
        a = math.pi*-5/8
        d = diameter * 1.5
        string_x = xhigh-trim + math.cos(a)*d
        string_z = zhigh      + math.sin(a)*d
        
        string_hole.move(string_x,0,string_z)
        
        #p = pad*0.5
        #loop = shape.Loop([(0,p),(p*2,-p),(-p*2,p)])
        #r = diameter-pad
        #stick = shape.extrusion([-r,r],[loop,loop])
        #
        #for i in xrange(-1,len(SCALE)+2):
        #    for j in xrange(int(zhigh / (diameter+pad))+1):
        #        x = (i-0.5)*(diameter+pad)
        #        z = r + j*(diameter+pad)
        #        c = stick.copy()
        #        if (i^j)&1:
        #            c.rotate(0,1,0,90)
        #        c.move(x,-diameter*0.5-pad,z)
        #        negatives.append(c)
        
        loop = shape.Loop([
            (xlow,0),
            (xhigh,0),
            (xhigh,zhigh-trim),
            (xhigh-trim,zhigh),
            (xlow,zhigh)
        ])

        bev = zsize / 4.0
        loop_bevel = shape.Loop([
            (xlow,bev),
            (xhigh,bev),
            (xhigh,zhigh-trim),
            (xhigh-trim,zhigh),
            (xlow,zhigh)
        ])
        
        #mask = loop.mask(RES)
        
        #amount = diameter * 0.5
        #op = shape.circle(amount).mask(RES)
        #mask = mask.open(op)
        
        #loop = mask.trace(RES)[0]
        
        #sloop = mask.erode(op).trace(RES)[0]
        
        #z1 = zsize*0.5-amount
        z2 = zsize*0.5
        
        #positive = shape.extrusion([-z2,z2],[loop,loop])
        positive = shape.extrusion([-z2,-z2+bev,z2-bev,z2],[loop_bevel,loop,loop,loop_bevel])
        positive.rotate(1,0,0,90)
        
        negative = string_hole.copy()
        for i, item in enumerate(negatives):
            print 'Merge %d / %d' % (i+1,len(negatives))
            negative.add(item)
        
        del negatives
        
        instrument = positive.copy()
        
        print 'Remove holes from instrument'
        instrument.remove(negative)
        
        del positive
        del negative
        
        instrument.rotate(1,0,0,90)
        extent = instrument.extent()

        copy = instrument.copy()
        copy.rotate(0,0,1,-45)
        cextent = copy.extent()
        copy.move(-(cextent.xmin+cextent.xmax)*0.5,
                  -(cextent.ymin+cextent.ymax)*0.5,
                  -cextent.zmin)
        self.save(copy,'instrument')
        del copy
        
        top = shape.block(
            extent.xmin-1,extent.xmax+1,
            extent.ymin-1,extent.ymax+1,
            0,extent.zmax+1
        )
        bottom = shape.block(
            extent.xmin-1,extent.xmax+1,
            extent.ymin-1,extent.ymax+1,
            extent.zmin-1,0
        )
        
        top.clip(instrument)
        bottom.clip(instrument)
        top.rotate(1,0,0,180)
        top.move(0,4,0)
        bottom.add(top)
        pattern = bottom
        
        del top
        del bottom
        
        pattern.move(0,0,z2)
        pattern.rotate(0,0,1, -90)

        bit_pad = 4.0
        extra_depth = 3.0
        a = bit_pad
        b = bit_pad + z2/10.0
        pad_cone = shape.extrude_profile(profile.Profile(
                [-extra_depth,0,z2],[a*2,a*2,b*2]
            ), 
            cross_section=lambda d: shape.circle(d,16))
        
        extra = b + 0.5
        
        extent = pattern.extent()
        mill = shape.block(
            extent.xmin-extra, extent.xmax+extra,
            extent.ymin-extra, extent.ymax+extra,
            -extra_depth, z2,
        )
        
        mill.remove( pattern.polygon_mask().to_3().minkowski_sum(pad_cone) )
        mill.add( pattern )
        
        del pad_cone
        del pattern
        
        self.save(mill, 'mill')
        
        
#        dilations = [
##            (z, shape.circle(2.0*( 2.0+z/20.0 )).mask(self.res))
##            for z in range(1)
#            (z, mask.ones(0,0,1,1) if not z else 
#                shape.circle(2.0*( z/10.0 )).mask(self.res))
#            for z in range(0,int(self.zsize),5)
#        ]

        #dilator = shape.circle(2* bit_diameter).mask(res)
        #hole_mask = mask.zero()
        #for x,y,packable in self.items:
        #    if packable.use_upper:
        #        hole_mask = hole_mask | packable.dilated_mask.offset(int(x*self.res+0.5),int(y*self.res+0.5))
        #hole_mask = hole_mask.close(dilator)
                
        #mask = bottom.mask(RES)
        #op = shape.circle(6).mask(RES)
        #loop = mask.dilate(op).trace(RES)
        #
        #z1 = -self.thickness+1.5
        #z2 = 0.0
        #loop_ext = shape.extrusion([z1,z2],[loop,loop])
        #extent = loop_ext.extent()
        #
        #cut = shape.block(
        #    extent.xmin-1,extent.xmax+1,
        #    extent.ymin-1,extent.ymax+1,
        #    z1,z2
        #)
        #cut.remove(loop_ext)
        #cut.add(bottom)
        #self.save(cut,'mill.stl')


if __name__ == '__main__': 
    shape.main_action(Make_panpipe())





import sys
from math import *

import fromcubic
import tocubic

import cornu

def eps_prologue(x0, y0, x1, y1, draw_box = False):
    print '%!PS-Adobe-3.0 EPSF'
    print '%%BoundingBox:', x0, y0, x1, y1
    print '%%EndComments'
    print '%%EndProlog'
    print '%%Page: 1 1'
    if draw_box:
        print x0, y0, 'moveto', x0, y1, 'lineto', x1, y1, 'lineto', x1, y0, 'lineto closepath stroke'

def eps_trailer():
    print '%%EOF'

def fit_cubic_superfast(z0, z1, arclen, th0, th1, aab):
    chord = hypot(z1[0] - z0[0], z1[1] - z0[1])
    cth0, sth0 = cos(th0), sin(th0)
    cth1, sth1 = -cos(th1), -sin(th1)
    armlen = .66667 * arclen
    a = armlen * aab
    b = armlen - a
    bz = [z0, (z0[0] + cth0 * a, z0[1] + sth0 * a),
          (z1[0] + cth1 * b, z1[1] + sth1 * b), z1]
    return bz

def fit_cubic(z0, z1, arclen, th_fn, fast, aabmin = 0, aabmax = 1.):
    chord = hypot(z1[0] - z0[0], z1[1] - z0[1])
    if (arclen < 1.000001 * chord):
        return [z0, z1], 0
    th0 = th_fn(0)
    th1 = th_fn(arclen)
    imax = 4
    jmax = 10
    if fast:
        imax = 1
        jmax = 0
    for i in range(imax):
        for j in range(jmax + 1):
            if jmax == 0:
                aab = 0.5 * (aabmin + aabmax)
            else:
                aab = aabmin + (aabmax - aabmin) * j / jmax
            if fast == 2:
                bz = fit_cubic_superfast(z0, z1, arclen, th0, th1, aab)
            else:
                bz = tocubic.fit_cubic_arclen(z0, z1, arclen, th0, th1, aab)
            score = tocubic.measure_bz_rk4(bz, arclen, th_fn)
            print '% aab =', aab, 'score =', score
            sys.stdout.flush()
            if j == 0 or score < best_score:
                best_score = score
                best_aab = aab
                best_bz = bz
        daab = .06 * (aabmax - aabmin)
        aabmin = max(0, best_aab - daab)
        aabmax = min(1, best_aab + daab)
        print '%--- best_aab =', best_aab
    return best_bz, best_score

def cornu_to_cubic(t0, t1, figno):
    if figno == 1:
        aabmin = 0
        aabmax = 0.4
    elif figno == 2:
        aabmin = 0.5
        aabmax = 1.
    else:
        aabmin = 0
        aabmax = 1.
    fast = 0
    if figno == 3:
        fast = 1
    elif figno == 4:
        fast = 2
    def th_fn(s):
        return (s + t0) ** 2
    y0, x0 = cornu.eval_cornu(t0)
    y1, x1 = cornu.eval_cornu(t1)
    bz, score = fit_cubic((x0, y0), (x1, y1), t1 - t0, th_fn, fast, aabmin, aabmax)
    return bz, score

def plot_k_of_bz(bz):
    dbz = tocubic.bz_deriv(bz)
    ddbz = tocubic.bz_deriv(dbz)
    cmd = 'moveto'
    ss = [0]
    def arclength_deriv(x, ss):
        dx, dy = tocubic.bz_eval(dbz, x)
        return [hypot(dx, dy)]
    dt = 0.01
    t = 0
    for i in range(101):
        dx, dy = tocubic.bz_eval(dbz, t)
        ddx, ddy = tocubic.bz_eval(ddbz, t)
        k = (ddy * dx - dy * ddx) / (dx * dx + dy * dy) ** 1.5
        print 100 + 500 * ss[0], 100 + 200 * k, cmd
        cmd = 'lineto'

        dsdx = arclength_deriv(t, ss)
        tocubic.rk4(ss, dsdx, t, .01, arclength_deriv)
        t += dt
    print 'stroke'

def plot_k_nominal(s0, s1):
    k0 = 2 * s0
    k1 = 2 * s1
    print 'gsave 0.5 setlinewidth'
    print 100, 100 + 200 * k0, 'moveto'
    print 100 + 500 * (s1 - s0), 100 + 200 * k1, 'lineto'
    print 'stroke grestore'

def simple_bez():
    eps_prologue(95, 126, 552, 508, 0)
    tocubic.plot_prolog()
    print '/ss 1.5 def'
    print '/circle { ss 0 moveto currentpoint exch ss sub exch ss 0 360 arc } bind def'
    bz, score = cornu_to_cubic(.5, 1.1, 2)
    fromcubic.plot_bzs([[bz]], (-400, 100), 1000, True)
    print 'stroke'
    print '/Times-Roman 12 selectfont'
    print '95 130 moveto ((x0, y0)) show'
    print '360 200 moveto ((x1, y1)) show'
    print '480 340 moveto ((x2, y2)) show'
    print '505 495 moveto ((x3, y3)) show'
    print 'showpage'
    eps_trailer()

def fast_bez(figno):
    if figno == 3:
        y1 = 520
    else:
        y1 = 550
    eps_prologue(95, 140, 552, y1, 0)
    tocubic.plot_prolog()
    print '/ss 1.5 def'
    print '/circle { ss 0 moveto currentpoint exch ss sub exch ss 0 360 arc } bind def'
    bz, score = cornu_to_cubic(.5, 1.1, figno)
    fromcubic.plot_bzs([[bz]], (-400, 100), 1000, True)
    print 'stroke'
    plot_k_nominal(.5, 1.1)
    plot_k_of_bz(bz)
    print 'showpage'
    eps_trailer()

def bezfig(s1):
    eps_prologue(95, 38, 510, 550, 0)
    #print '0.5 0.5 scale 500 100 translate'
    tocubic.plot_prolog()
    print '/ss 1.5 def'
    print '/circle { ss 0 moveto currentpoint exch ss sub exch ss 0 360 arc } bind def'
    bz, score = cornu_to_cubic(.5, 0.85, 1)
    fromcubic.plot_bzs([[bz]], (-400, 0), 1000, True)
    print 'stroke'
    plot_k_nominal(.5, 0.85)
    plot_k_of_bz(bz)
    bz, score = cornu_to_cubic(.5, 0.85, 2)
    fromcubic.plot_bzs([[bz]], (-400, 100), 1000, True)
    print 'stroke'
    print 'gsave 0 50 translate'
    plot_k_nominal(.5, .85)
    plot_k_of_bz(bz)
    print 'grestore'
    print 'showpage'

import sys

if __name__ == '__main__':
    figno = int(sys.argv[1])
    if figno == 0:
        simple_bez()
    elif figno == 1:
        bezfig(1.0)
    elif figno == 2:
        bezfig(0.85)
    else:
        fast_bez(figno)
    #fast_bez(4)



# A little solver for band-diagonal matrices. Based on NR Ch 2.4.

from math import *

from Numeric import *

do_pivot = True

def bandec(a, m1, m2):
    n, m = a.shape
    mm = m1 + m2 + 1
    if m != mm:
        raise ValueError('Array has width %d expected %d' % (m, mm))
    al = zeros((n, m1), Float)
    indx = zeros(n, Int)

    for i in range(m1):
        l = m1 - i
        for j in range(l, mm): a[i, j - l] = a[i, j]
        for j in range(mm - l, mm): a[i, j] = 0

    d = 1.

    l = m1
    for k in range(n):
        dum = a[k, 0]
        pivot = k
        if l < n: l += 1
        if do_pivot:
            for j in range(k + 1, l):
                if abs(a[j, 0]) > abs(dum):
                    dum = a[j, 0]
                    pivot = j
        indx[k] = pivot
        if dum == 0.: a[k, 0] = 1e-20
        if pivot != k:
            d = -d
            for j in range(mm):
                tmp = a[k, j]
                a[k, j] = a[pivot, j]
                a[pivot, j] = tmp
        for i in range(k + 1, l):
            dum = a[i, 0] / a[k, 0]
            al[k, i - k - 1] = dum
            for j in range(1, mm):
                a[i, j - 1] = a[i, j] - dum * a[k, j]
            a[i, mm - 1] = 0.
    return al, indx, d

def banbks(a, m1, m2, al, indx, b):
    n, m = a.shape
    mm = m1 + m2 + 1
    l = m1
    for k in range(n):
        i = indx[k]
        if i != k:
            tmp = b[k]
            b[k] = b[i]
            b[i] = tmp
        if l < n: l += 1
        for i in range(k + 1, l):
            b[i] -= al[k, i - k - 1] * b[k]
    l = 1
    for i in range(n - 1, -1, -1):
        dum = b[i]
        for k in range(1, l):
            dum -= a[i, k] * b[k + i]
        b[i] = dum / a[i, 0]
        if l < mm: l += 1

if __name__ == '__main__':
    a = zeros((10, 3), Float)
    for i in range(10):
        a[i, 0] = 1
        a[i, 1] = 2
        a[i, 2] = 1
    print a
    al, indx, d = bandec(a, 1, 1)
    print a
    print al
    print indx
    b = zeros(10, Float)
    b[5] = 1
    banbks(a, 1, 1, al, indx, b)
    print b



#!/usr/bin/env python

import demakein, nesoni

class Design_my_whistle(demakein.Design_folk_whistle):
    transpose = 12    
    
    # ... and any other things you want to change ...


if __name__ == '__main__':
    nesoni.run_toolbox(
        [ Design_my_whistle, demakein.Make_whistle ],
        show_make_flags=False)



"""

A basic flute or ney.

The embouchure hole is not modelled.
To use the design produced:
- make a flute without finger holes, just embouchure hole and end stopper.
- trim so that bottom note is in tune.
- drill finger holes.

The flute is a simple cylinder,
so the tuning is not perfect over the two registers
unless the finger holes are very large.

"""

import demakein, nesoni

class Flute(demakein.design.Instrument_designer):
    closed_top = False

    inner_diameters = [ 10.0, 10.0 ]
    outer_diameters = [ 14.0, 14.0 ]
    
    min_hole_diameters = [ 3.0 ]*6
    max_hole_diameters = [ 8.0 ]*6
    
    #min_hole_spacing = [ 15.0 ]*5
    #and/or
    balance = [ 0.05, None, None, 0.05 ]
    
    initial_length = demakein.design.wavelength('D5') * 0.5
    
    fingerings = [
        ('D5',  [1,1,1,1,1,1]),
        ('E5',  [0,1,1,1,1,1]),
        ('F#5', [0,0,1,1,1,1]),
        ('G5',  [0,0,0,1,1,1]),
        ('A5',  [0,0,0,0,1,1]),
        ('B5',  [0,0,0,0,0,1]),
        ('C#6', [0,0,0,0,0,0]),
        ('D6',  [1,1,1,1,1,1]),
        ('E6',  [0,1,1,1,1,1]),
        ('F#6', [0,0,1,1,1,1]),
        ('G6',  [0,0,0,1,1,1]),
        ('A6',  [0,0,0,0,1,1]),
        ('B6',  [0,0,0,0,0,1]),
        ('C#7', [0,0,0,0,0,0]),
        ('D8',  [1,1,1,1,1,1]),
        ]    

if __name__ == '__main__': 
    nesoni.run_tool(Flute)





import os

import profile, shape

def plan_segments(cuts, length_ratio):
    plans = [ ]
    for i in xrange(1<<len(cuts)):
        values = [ (i>>j)&1 for j in xrange(len(cuts)) ]
        upper = [0.0] + [ cuts[j] for j,k in enumerate(values) if k == 1] + [ 1.0 ]
        #lower = [0.0] + [ cuts[j] for j,k in enumerate(values) if k == 2] + [ 1.0 ]
        
        lower = [0.0] + [ (upper[j]+upper[j+1])*0.5 for j in xrange(len(upper)-1) ] + [1.0]
        j = 0
        while j < len(lower)-2:
            if lower[j+2]-lower[j] <= length_ratio:
                del lower[j+1]
            else:
                j += 1
        
        good = True
        for seg in (upper,lower):
            for j in xrange(len(seg)-1):
                if seg[j+1]-seg[j] > length_ratio:
                    good = False
                    break
            if not good: break
        if not good: continue
        plans.append(( upper, lower ))
    
    assert plans, 'Couldn\'t find a way to segment for milling.'    
    
    def scorer((upper,lower)):
        worst = 1e30
        for a in upper[1:-1]:
            for b in lower[1:-1]:
                worst = min(worst, abs(a-b))
        return len(upper)+len(lower),len(upper),-worst
    
    best = min(plans, key=scorer)
    print 'Segmentation:'
    print ' '.join('%3.2f' % i for i in best[0])
    print ' '.join('%3.2f' % i for i in best[1])
    print
    
    return best
                


class Packable(object):
    def __init__(self, shapes, rotation, dilation, use_upper=True):
        if not rotation:
            self.shapes = shapes
        else:
            self.shapes = [ item.copy() for item in shapes ]
            for item in self.shapes:
                item.rotate(0,0,1, rotation)
        
        extent = self.shapes[0].extent()
        for item in self.shapes:
            item.move(-extent.xmin,-extent.ymin,0)
        self.extent = self.shapes[0].extent()
        
        mask = self.shapes[0].polygon_mask()
        loop = mask.loop(holes=False)
        self.mask = shape.create_polygon_2(loop)
        self.centroid = loop.centroid

        self.dilated_mask = self.mask.offset_curve(dilation, 16)
        self.dilated_mask = shape.create_polygon_2(self.dilated_mask.loop(holes=False))

        self.dilated_extent = self.dilated_mask.extent()
        self.use_upper = use_upper


class Pack(object):
    def __init__(self, xsize, ysize, zsize, masks=[], items=[]):
        self.xsize = xsize
        self.ysize = ysize
        self.zsize = zsize
        self.masks = masks[:]
        self.items = items[:] #(x,y,packable)

    def copy(self):
        return Pack(self.xsize, self.ysize, self.zsize, self.masks, self.items)

    def put(self, x,y,packable):
        self.items.append((x,y,packable))
        shifted = packable.mask.copy()
        shifted.move(x,y)
        self.masks.append(shifted)
    
    def valid(self, x,y,packable):
        if (packable.extent.zmax > self.zsize or 
            x+packable.dilated_extent.xmin < 0.0 or
            y+packable.dilated_extent.ymin < 0.0 or
            x+packable.dilated_extent.xmax > self.xsize or
            y+packable.dilated_extent.ymax > self.ysize):
            return False
        shifted = packable.dilated_mask.copy()
        shifted.move(x,y)
        for mask in self.masks:
            if mask.intersects(shifted):
                return False
        return True

    def render(self, bit_diameter, bit_pad):
        xsize = self.xsize
        ysize = self.ysize
        
        a = bit_pad
        b = bit_pad + self.zsize/10.0
        pad_cone = shape.extrude_profile(profile.Profile([0,self.zsize],[a*2,b*2],[a*2,b*2]), cross_section=lambda d: shape.circle(d,16))
        pad_cylinder = shape.extrude_profile(profile.Profile([0,self.zsize+bit_diameter],[a*2,a*2],[a*2,a*2]), cross_section=lambda d: shape.circle(d,16))
        
        pad = 0.5 + self.zsize/10.0
        
        upper = shape.block(-pad,xsize+pad,-pad,ysize+pad,0,self.zsize)
        
        print 'Cut holes'

        for i,(x,y,packable) in enumerate(self.items):
            if packable.use_upper:
                flat = packable.mask.to_3()
                flat.move(x,y,0)
                minsum = flat.minkowski_sum(pad_cone)
                upper.remove( minsum )

        print 'Put things in them'

        for x,y,packable in self.items:
            if packable.use_upper:
                temp = packable.shapes[0].copy()
                temp.move(x,y,0)
                upper.add(temp)
            
        bol = int(self.xsize / 10.0)
        bol_width = 1.0
            # 0.5 seemed to allow shifting on a sop flute
        bol_height = 1.0 
            # 0.5 was too low, middle piece tore loose in final pass of
            # contour finishing (x-scan of horizontal-like surfaces)
        xmins = [ x for x,y,packable in self.items ]
        xmaxs = [ x+packable.extent.xmax for x,y,packable in self.items ]
        margin = bol_width+bit_diameter+0.25
        
        potential_bols = [ ((i+0.5)*xsize/bol, 0,ysize) for i in xrange(bol) ]
        bols = [ ]
        while potential_bols:
            pos, y_low, y_high = potential_bols.pop()
            for i, (x,y,packable) in enumerate(self.items):
                if pos-margin <= xmins[i] <= pos+margin or \
                   pos-margin <= xmaxs[i] <= pos+margin:
                   low = y-bit_diameter
                   high = y+packable.extent.ymax+bit_diameter
                   if low <= y_low and y_high <= high:
                       break
                   elif y_low < low and high < y_high:
                       potential_bols.append((pos, y_low, low))
                       potential_bols.append((pos, high, y_high))
                       break
                   elif y_low < low < y_high <= high:
                       potential_bols.append((pos, y_low, low))
                       break
                   elif low < y_low < high <= y_high:
                       potential_bols.append((pos, high, y_high))
                       break
            else:
                bols.append((pos,y_low,y_high))
                
        for pos, y_low, y_high in bols:    
            upper.add(shape.block(
                pos-bol_width*0.5, pos+bol_width*0.5,
                y_low, y_high,
                #0,ysize,
                #-bit_diameter*0.6, bol_height,
                0,bol_height,
            ))     
        
        
        print 'Underside'
        
        # Cut the bottom of upper with lower, to depth    bit_diameter
        lower = shape.block(-pad,xsize+pad,-pad,ysize+pad,bit_diameter,self.zsize)
        lower.add(upper)
        
        print 'Remove bores'

        for x,y,packable in self.items:
            temp = packable.shapes[1].copy()
            temp.move(x,y,0)

            flat = packable.mask.to_3()
            flat.move(x,y,0)
            minsum = flat.minkowski_sum(pad_cylinder)
            temp.clip(minsum)

            lower.remove(temp)
        
        lower.rotate(0,1,0, 180)
        
        return lower, upper
    
    def render_print(self):
        result = None
        for x,y,packable in self.items:
            temp = packable.shapes[0].copy()
            temp.move(x,y,0)
            if result is None:
                result = temp
            else:
                result.add(temp)
        return result


def make_segment(instrument, top, low, high, radius, clip=True):
    #if not clip_half:
    #   y1,y2 = -radius,radius
    segment = instrument.copy()
    
    if clip:
        if top:
           y1,y2 = 0,radius
        else:
           y1,y2 = -radius,0
        clipper = shape.block(-radius,radius, y1,y2, low,high)
        segment.clip(clipper)
    #Closed flute is meant to be like this
    #assert abs(segment.size()[2] - (high-low+pad*2)) < 1e-3, 'Need more padding for construction'
    segment.move(0,0,-low)
    if top:
       segment.rotate(0,0,1, 180)

    segment.rotate(1,0,0,-90)
    segment.rotate(0,0,1,90)
    segment.move(high-low,0,0)
    return segment


def make_segments(instrument, length, radius, top_fractions, bottom_fractions, clip=True):
    parts = [ ]
    z = [ item*length for item in top_fractions ]
    for i in range(len(top_fractions)-1):
        parts.append(make_segment(instrument, True, 
                     z[i],z[i+1], 
                     radius, clip))

    z = [ item*length for item in bottom_fractions ]
    for i in range(len(bottom_fractions)-1):    
        parts.append(make_segment(instrument, False, 
                     z[i],z[i+1], 
                     radius, clip))

    return parts


def deconstruct(outer, bore, top_fractions, bottom_fractions, 
                bit_diameter, dilation, block_zsize):
    """ Object must be large enough to include end padding """
    xsize,ysize,zsize = outer.size()
    radius = max(xsize,ysize)*2
    
    length = zsize
    
    outers = make_segments(outer, length, radius, top_fractions, bottom_fractions)
    bores  = make_segments(bore, length, radius, top_fractions, bottom_fractions, False)

    result = [ ]
    for outer, bore in zip(outers, bores):
        extent = outer.extent()
        
        b = bit_diameter
        clipper = shape.block(extent.xmin-b,extent.xmax+b,
                              extent.ymin-b,extent.ymax+b,
                              extent.zmin,extent.zmax+b)
        bore.clip(clipper)
        
        n = 1
        while extent.zmax > block_zsize*n:
            n += 1
        #thickness = extent.zmax/n
        thickness = block_zsize
        
        block = shape.block(extent.xmin-1,extent.xmax+1,extent.ymin-1,extent.ymax+1,0,thickness)
        for i in xrange(n):
            temp_outer = outer.copy()
            temp_bore = bore.copy()
            temp_outer.move(0,0,-i*thickness)
            temp_bore.move(0,0,-i*thickness)
            temp_outer.clip(block)
            #temp_bore.clip(block)
            result.append([
                Packable([temp_outer,temp_bore], 0, dilation),
                Packable([temp_outer,temp_bore], 180, dilation),
                ])
    return result

    #return [ [ Packable([outer, bore], 0, dilation),
    #            Packable([outer, bore], 180, dilation) ]
    #          for outer, bore in zip(outers, bores) ]


def mill_template(xsize, ysize, zsize, dilation):
    template = Pack(xsize, ysize, zsize)
    
    peg_diameter = 6.0
    block_radius = 4.0
    hole = shape.prism(min(zsize,peg_diameter*2.0), peg_diameter)   # *3 is too deep
    hole_block = shape.block(-block_radius,block_radius,-block_radius,block_radius,0,zsize)
    hole_packable = Packable([hole_block,hole], 0, 0.0, use_upper=False)
    
    shift = 4
    template.put(shift, block_radius,hole_packable)
    template.put(xsize-block_radius*2-shift, block_radius,hole_packable)

    template.put(0,ysize-block_radius*3,hole_packable)
    template.put(xsize-block_radius*2,ysize-block_radius*3,hole_packable)
    
    return template

def mill_squish(pack):
    x1, y, hole1 = pack.items.pop(2)
    x2, y, hole2 = pack.items.pop(2)
    del pack.masks[2:4]
    y_new = y
    step = 1.0
        
    y_min = max( item[1]+item[2].extent.ymax for item in pack.items )
    
    while y_new-step >= y_min and pack.valid(x1,y_new-step,hole1) and pack.valid(x2,y_new-step,hole2):
        y_new -= step
        
    pack.put(x1,y_new,hole1)
    pack.put(x2,y_new,hole2)
    pack.ysize -= y-y_new


def pack(template, packables, aspect_goal=1.0):    
    packs = [ template.copy() ]
    pack = packs[-1]
    
    step = 2.0
    step_back = 0.1

    points = [ ]
    for y in xrange(int(template.ysize/step)):
        for x in xrange(int(template.xsize/step)):
            points.append((x*step,y*step))
    
    todo = list(packables)
    while todo:
        try:
            queue = [
                (i,j,x,y)
                for x,y in points
                for i in xrange(len(todo))
                for j in xrange(len(todo[i]))
            ]
        
            #for i in range(len(todo)):
            #    for x,y in points:
            
            def sorter(item):
                i,j,x,y = item
                xm = todo[i][j].extent.xmax
                ym = todo[i][j].extent.ymax
                xc, yc = todo[i][j].centroid
                return (((aspect_goal*(x+xc))**2+(y+yc)**2) - ((aspect_goal*xm)**2+(ym)**2), y, x)
                #return (max(aspect_goal*(x+xc),(y+yc)) -max(aspect_goal*xc,yc), y, x)
                #return (
                #    y > 0,
                #    -todo[i][0].extent.xmax,
                #    y,
                #    x,
                #)
            queue.sort(key=sorter)
            
            for i,j,x,y in queue:
                if pack.valid(x,y,todo[i][j]):
                    raise StopIteration()

            assert len(pack.items) > len(template.items), 'This is never going to work'
            pack = template.copy()
            packs.append(pack)
            
        except StopIteration:
            while True:
                if pack.valid(x-step_back,y,todo[i][j]):
                    x -= step_back
                elif pack.valid(x,y-step_back,todo[i][j]):
                    y -= step_back
                else:
                    break
            
            print len(todo),'put',len(packs),x,y
            
            pack.put(x,y,todo[i][j])
            del todo[i]

    return packs


def save_packs(packs, save, bit_diameter, pad):
    shapes = [ ]
    for i, item in enumerate(packs):
        lower, upper = item.render(bit_diameter, pad)
        shapes.extend([lower,upper])
        save(lower, 'lower-%d-of-%d' % (i+1,len(packs)))
        save(upper, 'upper-%d-of-%d' % (i+1,len(packs)))

    return shapes


def cut_and_pack(outer, bore, top_fractions, bottom_fractions, xsize, ysize, zsize, bit_diameter, save, extra_packables=[]):
    pad = bit_diameter*1.2
    dilation = pad
    
    packables = deconstruct(outer, bore, top_fractions, bottom_fractions, bit_diameter, dilation, zsize)
    packables.extend(extra_packables)

    for item in packables:
        print '%.1fmm x %.1fmm x %.1fmm' % (item[0].extent.xmax,item[0].extent.ymax,item[0].extent.zmax)

    print 'Packing'

    template = mill_template(xsize + pad*2, ysize + pad*2, zsize, dilation)
    packs = pack(template, packables, 0.5*float(ysize)/xsize)
    for item in packs:
        mill_squish(item)
    
    print 'Rendering'
    
    return save_packs(packs, save, bit_diameter, pad)






# Convert piecewise cubic into piecewise clothoid representation.

from math import *

import clothoid
import pcorn
import tocubic

import offset

def read_bz(f):
    result = []
    for l in f.xreadlines():
        s = l.split()
        if len(s) > 0:
            cmd = s[-1]
            #print s[:-1], cmd
            if cmd == 'm':
                sp = []
                result.append(sp)
                curpt = [float(x) for x in s[0:2]]
                startpt = curpt
            elif cmd == 'l':
                newpt = [float(x) for x in s[0:2]]
                sp.append((curpt, newpt))
                curpt = newpt
            elif cmd == 'c':
                c1 = [float(x) for x in s[0:2]]
                c2 = [float(x) for x in s[2:4]]
                newpt = [float(x) for x in s[4:6]]
                sp.append((curpt, c1, c2, newpt))
                curpt = newpt
    return result

def plot_bzs(bzs, z0, scale, fancy = False):
    for sp in bzs:
        for i in range(len(sp)):
            bz = sp[i]
            tocubic.plot_bz(bz, z0, scale, i == 0)
        print 'stroke'
        if fancy:
            for i in range(len(sp)):
                bz = sp[i]

                x0, y0 = z0[0] + scale * bz[0][0], z0[1] + scale * bz[0][1]
                print 'gsave', x0, y0, 'translate circle fill grestore'
                if len(bz) == 4:
                    x1, y1 = z0[0] + scale * bz[1][0], z0[1] + scale * bz[1][1]
                    x2, y2 = z0[0] + scale * bz[2][0], z0[1] + scale * bz[2][1]
                    x3, y3 = z0[0] + scale * bz[3][0], z0[1] + scale * bz[3][1]
                    print 'gsave 0.5 setlinewidth', x0, y0, 'moveto'
                    print x1, y1, 'lineto stroke'
                    print x2, y2, 'moveto'
                    print x3, y3, 'lineto stroke grestore'
                    print 'gsave', x1, y1, 'translate 0.75 dup scale circle fill grestore'
                    print 'gsave', x2, y2, 'translate 0.75 dup scale circle fill grestore'
                    print 'gsave', x3, y3, 'translate 0.75 dup scale circle fill grestore'
            
        

def measure_bz_cloth(seg, bz, n = 100):
    bz_arclen = tocubic.bz_arclength_rk4(bz)
    arclen_ratio = seg.arclen / bz_arclen
    dbz = tocubic.bz_deriv(bz)
    
    def measure_derivs(x, ys):
        dx, dy = tocubic.bz_eval(dbz, x)
        ds = hypot(dx, dy)
        s = ys[0] * arclen_ratio
        dscore = ds * (tocubic.mod_2pi(atan2(dy, dx) - seg.th(s)) ** 2)
        #print s, atan2(dy, dx),  seg.th(s)
        return [ds, dscore]
    dt = 1./n
    t = 0
    ys = [0, 0]
    for i in range(n):
        dydx = measure_derivs(t, ys)
        tocubic.rk4(ys, dydx, t, dt, measure_derivs)
        t += dt
    return ys[1]

def cubic_bz_to_pcorn(bz, thresh):
    dx = bz[3][0] - bz[0][0]
    dy = bz[3][1] - bz[0][1]
    dx1 = bz[1][0] - bz[0][0]
    dy1 = bz[1][1] - bz[0][1]
    dx2 = bz[3][0] - bz[2][0]
    dy2 = bz[3][1] - bz[2][1]
    chth = atan2(dy, dx)
    th0 = tocubic.mod_2pi(chth - atan2(dy1, dx1))
    th1 = tocubic.mod_2pi(atan2(dy2, dx2) - chth)
    seg = pcorn.Segment(bz[0], bz[3], th0, th1)
    err = measure_bz_cloth(seg, bz)
    if err < thresh:
        return [seg]
    else:
        # de Casteljau
        x01, y01 = 0.5 * (bz[0][0] + bz[1][0]), 0.5 * (bz[0][1] + bz[1][1])
        x12, y12 = 0.5 * (bz[1][0] + bz[2][0]), 0.5 * (bz[1][1] + bz[2][1])
        x23, y23 = 0.5 * (bz[2][0] + bz[3][0]), 0.5 * (bz[2][1] + bz[3][1])
        xl2, yl2 = 0.5 * (x01 + x12), 0.5 * (y01 + y12)
        xr1, yr1 = 0.5 * (x12 + x23), 0.5 * (y12 + y23)
        xm, ym = 0.5 * (xl2 + xr1), 0.5 * (yl2 + yr1)
        bzl = [bz[0], (x01, y01), (xl2, yl2), (xm, ym)]
        bzr = [(xm, ym), (xr1, yr1), (x23, y23), bz[3]]
        segs = cubic_bz_to_pcorn(bzl, 0.5 * thresh)
        segs.extend(cubic_bz_to_pcorn(bzr, 0.5 * thresh))
        return segs

def bzs_to_pcorn(bzs, thresh = 1e-9):
    result = []
    for sp in bzs:
        rsp = []
        for bz in sp:
            if len(bz) == 2:
                dx = bz[1][0] - bz[0][0]
                dy = bz[1][1] - bz[0][1]
                th = atan2(dy, dx)
                rsp.append(pcorn.Segment(bz[0], bz[1], 0, 0))
            else:
                rsp.extend(cubic_bz_to_pcorn(bz, thresh))
        result.append(rsp)
    return result

def plot_segs(segs):
    for i in range(len(segs)):
        seg = segs[i]
        if i == 0:
            print seg.z0[0], seg.z0[1], 'moveto'
        print seg.z1[0], seg.z1[1], 'lineto'
    print 'stroke'
    for i in range(len(segs)):
        seg = segs[i]
        if i == 0:
            print 'gsave', seg.z0[0], seg.z0[1], 'translate circle fill grestore'
        print 'gsave', seg.z1[0], seg.z1[1], 'translate circle fill grestore'

import sys

def test_to_pcorn():
    C1 = 0.55228
    bz = [(100, 100), (100 + 400 * C1, 100), (500, 500 - 400 * C1), (500, 500)]
    for i in range(0, 13):
        thresh = .1 ** i
        segs = cubic_bz_to_pcorn(bz, thresh)
        plot_segs(segs)
        print >> sys.stderr, thresh, len(segs)
        print '0 20 translate'

if __name__ == '__main__':
    f = file(sys.argv[1])
    bzs = read_bz(f)
    rsps = bzs_to_pcorn(bzs, 1)
    #print rsps
    tocubic.plot_prolog()
    print 'grestore'
    print '1 -1 scale 0 -720 translate'
    print '/ss 1.5 def'
    print '/circle { ss 0 moveto currentpoint exch ss sub exch ss 0 360 arc } bind def'
    tot = 0
    for segs in rsps:
        curve = pcorn.Curve(segs)
        #curve = offset.offset(curve, 10)
        print '%', curve.arclen
        print '%', curve.sstarts
        if 0:
            print 'gsave 1 0 0 setrgbcolor'
            cmd = 'moveto'
            for i in range(100):
                s = i * .01 * curve.arclen
                x, y = curve.xy(s)
                th = curve.th(s)
                sth = 5 * sin(th)
                cth = 5 * cos(th)
                print x, y, cmd
                cmd = 'lineto'
            print 'closepath stroke grestore'
        for i in range(100):
            s = i * .01 * curve.arclen
            x, y = curve.xy(s)
            th = curve.th(s)
            sth = 5 * sin(th)
            cth = 5 * cos(th)
            if 0:
                print x - cth, y - sth, 'moveto'
                print x + cth, y + sth, 'lineto stroke'
        if 1:
            for s in curve.find_breaks():
                print 'gsave 0 1 0 setrgbcolor'
                x, y = curve.xy(s)
                print x, y, 'translate 2 dup scale circle fill'
                print 'grestore'
        #plot_segs(segs)

        print 'gsave 0 0 0 setrgbcolor'
        optim = 3
        thresh = 1e-2
        new_bzs = tocubic.pcorn_curve_to_bzs(curve, optim, thresh)
        tot += len(new_bzs)
        plot_bzs([new_bzs], (0, 0), 1, True)
        print 'grestore'
    print 'grestore'
    print '/Helvetica 12 selectfont'
    print '36 720 moveto (thresh=%g optim=%d) show' % (thresh, optim)
    print '( tot segs=%d) show' % tot
    print 'showpage'

    #plot_bzs(bzs, (100, 100), 1)




import sys, collections, math

import shape, svg

def sub(a,b):
    return tuple( aa-bb for aa,bb in zip(a,b) )

def dot(a,b):
    return sum( aa*bb for aa,bb in zip(a,b) )

def cross(a,b):
    return (a[1]*b[2]-a[2]*b[1], a[2]*b[0]-a[0]*b[2], a[0]*b[1]-a[1]*b[0])

def sketch(thing, outname):
    print outname,
    sys.stdout.flush()
    
    ox = 0
    oy = 0
    
    pic = svg.SVG()
    pic.require(0,0)
    
    triangles = thing.triangles()
    
    def rot(triangles, x,y,z,angle):
        mat = shape.rotation_matrix(x,y,z,angle)
        return [ [ shape.transform_point_3(mat,item2) for item2 in item ] for item in triangles ]

    def iterator():
        #yield thing, 1
        yield triangles, 1
        
        #item_rot = thing.copy()
        #item_rot.rotate(1,0,0,-90)
        #yield item_rot, 1
        #del item_rot
        yield rot(triangles, 1,0,0,-90), 1
        
        #item_iso = thing.copy()
        #item_iso.rotate(0,0,1,-45, 256)
        #item_iso.rotate(1,0,0,-45, 256)
        #yield item_iso, 0
        #del item_iso        
        yield rot(rot(triangles, 0,0,1,-45), 1,0,0,-45), 0
    
        
#    for (item,showdim) in iterator():
#        extent = item.extent()
    for (this_triangles,showdim) in iterator():
        extent = shape.extent_3(item2 for item in this_triangles for item2 in item)
        
        if extent.xmax-extent.xmin < extent.ymax-extent.ymin:
            ox = pic.max_x + 10
            oy = -pic.max_y
        else:
            ox = 0.0
            oy = -(extent.ymax-extent.ymin)-pic.max_y - 20
        
        ox -= extent.xmin
        oy -= extent.ymin
        
        xmid = (extent.xmin+extent.xmax)*0.5
        ymid = (extent.ymin+extent.ymax)*0.5
        zmid = (extent.zmin+extent.zmax)*0.5
        if showdim:
            pic.text(ox+xmid, -oy-extent.ymax-10, '%.1fmm' % (extent.xmax-extent.xmin))
            pic.text(ox+extent.xmax + 5, -oy-ymid, '%.1fmm' % (extent.ymax-extent.ymin))    
        
        lines = collections.defaultdict(list)
        for tri in this_triangles:
            #a = numpy.array(tri)
            #normal = numpy.cross(a[1]-a[0],a[2]-a[0])
            normal = cross(sub(tri[1],tri[0]),sub(tri[2],tri[0]))
            length = math.sqrt(normal[0]**2+normal[1]**2+normal[2]**2)
            normal = (normal[0]/length,normal[1]/length,normal[2]/length)
            
            lines[(tri[0],tri[1])].append( normal )
            lines[(tri[1],tri[2])].append( normal )
            lines[(tri[2],tri[0])].append( normal )
    
        for a,b in lines:
            if (b,a) not in lines:
                weight = -1.0
                normals = lines[(a,b)]
            else:
                if (b,a) < (a,b): continue
                normals = lines[(a,b)] + lines[(b,a)]
                
                weight = -1.0
                for n1 in lines[(a,b)]:
                    for n2 in lines[(b,a)]:
                        if n1[2]*n2[2] > 0.0:
                            weight = max(weight, dot(n1,n2))
            
            weight = (1.0-weight)*0.5                
            if weight < 0.01: continue                
            weight = min(1.0,weight*2)
    
            outs = 0
            ins = 0
            for n in normals:
                if n[2] <= 1e-6:
                    ins += 1
                if n[2] >= -1e-6:
                    outs += 1
            
            if ins >= 2 and not outs:
                weight *= 0.25
            #elif outs >= 2 and not ins:
            #    weight *= 2.0
            
            pic.line([ (ox+x,-oy-y) for x,y,z in [a,b] ],width=0.2 * weight)
        
        del lines
        
    pic.save(outname)
    print


def run(args):
    for filename in args:
        prefix = filename    
        if prefix[-4:].lower() == '.stl': 
            prefix = prefix[:-4]
        outname = prefix + '-sketch.svg'
    
        item = shape.load_stl(filename)

        sketch(item, outname)
        
   
   
if __name__ == '__main__':
    shape.main(run)